#include "proj.h"
#define DEBUG
#ifndef DEBUG
  #define eprintf( fmt, args... )
  #define DBGOUT(...)
#else
  #if DEBUGHW>0
    #define FOO(...) __VA_ARGS__
    #define DBGOUT dbg_string+= FOO
    #if (DEBUGHW==2)
      #define eprintf(fmt, args...) S.printf(fmt, ##args)
    #elif (DEBUGHW==1 || DEBUGHW==3)
      #define eprintf(fmt, args...) sprintf(dbg,fmt, ##args);dbg_string+=dbg;dbg[0]=0
    #endif
  #else
    #define eprintf( fmt, args... )
    #define DBGOUT(...)
  #endif
#endif

String  printIP(IPAddress adress) {
  return (String)adress[0] + "." + (String)adress[1] + "." + (String)adress[2] + "." + (String)adress[3];
}

void evseCommand(uint32_t clientId) {
}

void  sendEventLog(uint32_t clientId,int page) {
  DynamicJsonBuffer jsonBuffer;
  JsonObject &root = jsonBuffer.createObject();
  root["page"] = page;
  JsonArray &items = root.createNestedArray("list");
  File eventlog = SPIFFS.open("/eventlog.json", "r");
  int first = (page - 1) * 20;
  int last = page * 20;
  int i = 0;
  while(eventlog.available()) {
    String item = String();
    item = eventlog.readStringUntil('\n');
    if(i >= first && i < last) {
      items.add(item);
    }
    i++;
  }
  eventlog.close();
  float pages = i / 20.0;
  root["haspages"] = ceil(pages);
  String buffer;
  root.prettyPrintTo(buffer);
  ws.text(clientId,buffer);
}

void sendZDataWait() {
  DynamicJsonBuffer jsonBuffer;
  JsonObject &doc = jsonBuffer.createObject();
  doc["now"] = 0;
  doc["uptime"] = millis()/1000;
//  doc["things_up"] = things_up;
  size_t len = doc.measureLength();
  AsyncWebSocketMessageBuffer *buffer = ws.makeBuffer(len);
  if(buffer) {
    doc.printTo((char *)buffer->get(), len + 1);
    ws.textAll(buffer);
  }
}

void sendZData() {
  DynamicJsonBuffer jsonBuffer;
  JsonObject &doc = jsonBuffer.createObject();
  doc["now"] = timecode;
  doc["1_8_0"] = a_result[0];
  doc["2_8_0"] = a_result[1];
  doc["3_8_1"] = a_result[2];
  doc["4_8_1"] = a_result[3];
  doc["1_7_0"] = a_result[4];
  doc["2_7_0"] = a_result[5];
  doc["3_7_0"] = a_result[6];
  doc["4_7_0"] = a_result[7];
  doc["1_128_0"] = a_result[8];
  doc["uptime"] = millis()/1000;
  doc["things_up"] = things_up;

  size_t len = doc.measureLength();
  AsyncWebSocketMessageBuffer *buffer = ws.makeBuffer(len);
  if(buffer) {
    doc.printTo((char *)buffer->get(), len + 1);
    ws.textAll(buffer);
  }
}

void printScanResult(int nFound) {
  String Networks=F("{\"scan_wifi\":\"");
  Networks += String(nFound)+F(" Netzwerk");
  if (nFound>1) Networks += "e";
  Networks += F(" gefunden<br><table class=\\\"pure-table\\\">");
  Networks += F("<thead><tr><th>Name (SSID)</th><th>Signal</th><th>Verschlüsselt</th></tr></thead><tbody>");
  for(int i = 0; i < nFound; ++i) {
    Networks += "<tr><td>"  +  String(WiFi.SSID(i))  + "</td><td>" +  String(WiFi.RSSI(i)) + " dB</td><td>" +  String((WiFi.encryptionType(i) == ENC_TYPE_NONE)?" ":"Ja")  + "</td></tr>";
  }
  Networks += F("</tbody></table>\"}");

  size_t len = Networks.length();
  AsyncWebSocketMessageBuffer *buffer = ws.makeBuffer(len); //  creates a buffer (len + 1) for you.
  if(buffer) {
    strcpy((char*)buffer->get(),Networks.c_str());
    ws.textAll(buffer);
  }
  WiFi.scanDelete();
}

void  sendStatus(uint32_t clientId) {
  struct ip_info info;
  FSInfo fsinfo;
  if(!SPIFFS.info(fsinfo)) {
    DBGOUT(F("[ WARN ] Error getting info on SPIFFS"));
  }
  DynamicJsonBuffer jsonBuffer;
  JsonObject &root = jsonBuffer.createObject();
  //root["command"] = "status";
  root["chipid"] = String(ESP.getChipId(), HEX);

  root["heap"] = ESP.getFreeHeap();
  root["cpu"] = ESP.getCpuFreqMHz();
  root["sketchsize"] = ESP.getSketchSize();
  root["freesize"] = ESP.getFreeSketchSpace();
  root["availspiffs"] = fsinfo.totalBytes - fsinfo.usedBytes;
  root["spiffssize"] = fsinfo.totalBytes;
  root["version"] = VERSION;

  if(inAPMode) {
    wifi_get_ip_info(SOFTAP_IF, &info);
    struct softap_config conf;
    wifi_softap_get_config(&conf);
//    root["ssid"] = String(reinterpret_cast<char *>(conf.ssid));
//    root["dns"] = printIP(WiFi.softAPIP());
    root["mac"] = WiFi.softAPmacAddress();
  }
  else {
    wifi_get_ip_info(STATION_IF, &info);
    struct station_config conf;
    wifi_station_get_config(&conf);
    root["ssid"] = String(reinterpret_cast<char *>(conf.ssid));
    root["dns"] = printIP(WiFi.dnsIP());
    root["mac"] = WiFi.macAddress();
    root["channel"] = WiFi.channel();
    root["rssi"] = WiFi.RSSI();
  }
  IPAddress ipaddr = IPAddress(info.ip.addr);
  IPAddress gwaddr = IPAddress(info.gw.addr);
  IPAddress nmaddr = IPAddress(info.netmask.addr);
  root["deviceip"] = printIP(ipaddr);
  root["gateway"] = printIP(gwaddr);
  root["netmask"] = printIP(nmaddr);
  root["heap"] = ESP.getFreeHeap();
  //root["loadaverage"] = systemLoadAverage();
  //if (ADC_MODE_VALUE == ADC_VCC) {
  root["vcc"] = ESP.getVcc();
  //root["vcc"] = "N/A (TOUT) ";
  root["sdk"] = ESP.getSdkVersion();
  root["core"] = ESP.getCoreVersion();
  root["mqttStatus"] = mqttStatus ? "connected":"N/A";
  root["ntpSynced"] = "N/A";
  String buffer;
  root.printTo(buffer);
  ws.text(clientId,buffer);
}

void send_data_file(const char *filename, uint32_t clientId) {
  DBGOUT("send file: "+String(filename)+'\n');
  File configFile = SPIFFS.open(filename, "r");
  if(configFile) {
    ws.text(clientId,configFile.readString());
    configFile.close();
  }
  else eprintf("File %s not found\n",filename);
}

void  wsClientRequest(AsyncWebSocketClient *client, size_t sz) {
  DynamicJsonBuffer jsonBuffer;
  JsonObject &root = jsonBuffer.parseObject((char *)(client->_tempObject));
  if(!root.success()) {
    DBGOUT(F("[ WARN ] Couldn't parse WebSocket message"));
    free(client->_tempObject);
    client->_tempObject = NULL;
    return;
  }
  // Web Browser sends some commands, check which command is given
  clientId=client->id();
  const char *command = root["command"];
  if(strcmp(command, "ping") != 0)
    eprintf("[ INFO ] command: %s\n",command);
  // Check whatever the command is and act accordingly
  if((strcmp(command, "/config_general")==0) || (strcmp(command, "/config_wifi")==0) || (strcmp(command, "/config_mqtt")==0)) {
    File f = SPIFFS.open(command, "w+");
    if(f) {
      size_t len = root.measurePrettyLength();
      root.prettyPrintTo(f);
      //root.prettyPrintTo(dbg_string);
      f.close();
      eprintf("[ INFO ] %s stored in the SPIFFS (%u bytes)\n",command,len);
    }
  }
  else if(strcmp(command, "status") == 0) {
    sendStatus(client->id());
  }
  else if(strcmp(command, "ping") == 0) {
    ws.text(client->id(),F("{\"pong\":\"\"}"));
  }
  else if(strcmp(command, "restart") == 0) {
    shouldReboot = true;
  }
  else if(strcmp(command, "geteventlog") == 0) {
    logPage = root["page"];
    //Log in main.loop bearbeiten, Timeout!!!
  }
  else if(strcmp(command, "clearevent") == 0) {
    SPIFFS.remove("/eventlog.json");
    writeEvent("WARN", "sys", "Event log cleared!", "");
  }
  else if(strcmp(command, "scan_wifi") == 0) {
    WiFi.scanNetworksAsync(printScanResult, true);
  }
  else if(strcmp(command, "getconf") == 0) {
    send_data_file("/config_general",client->id());
    send_data_file("/config_wifi",client->id());
    send_data_file("/config_mqtt",client->id());
    histUpdate();
  }
  else if(strcmp(command, "clearhist") == 0) {
    String s="";
    for (unsigned j=0; j<7;j++ ) {
      kwh_day_in[j]=0;
      kwh_day_out[j]=0;
      kwh_hist[j].kwh_in=0;
      kwh_hist[j].kwh_out=0;
      kwh_hist[j].dow=0;
      s="/hist"+String(j);
      SPIFFS.remove(s);
    }
    first_frame=0;
  }
  else if(strcmp(command, "ls") == 0) {
    DynamicJsonBuffer jsonBuffer;
    JsonObject &doc = jsonBuffer.createObject();
    Dir dir = SPIFFS.openDir("/");
    unsigned i=0;
    while (dir.next()) {
      File f = dir.openFile("r");
      //eprintf("%s \t %u\n",dir.fileName().c_str(),f.size());
      doc[String(i)]=dir.fileName()+' '+String(f.size());
      i++;
    }
    doc["ls"]=i;
    String buffer;
    doc.printTo(buffer);
    //DBGOUT(buffer+"\n");
    ws.text(clientId,buffer);
//    size_t len = doc.measureLength();
//    AsyncWebSocketMessageBuffer *buffer = ws.makeBuffer(len);
//    if(buffer) {
//      doc.printTo((char *)buffer->get(), len + 1);
//      ws.textAll(buffer);
//    }
  }
  else if(strcmp(command, "clear") == 0) {
    SPIFFS.remove(F("/config_general"));
    SPIFFS.remove(F("/config_wifi"));
    SPIFFS.remove(F("/config_mqtt"));
  }
  else if(strcmp(command, "test") == 0) hwTest=true;
  else if(strcmp(command, "remove") == 0) {
    const char *uid = root["file"];
    SPIFFS.remove(uid);
  }

  free(client->_tempObject);
  client->_tempObject = NULL;
}
